# these need to be at the top level:

# A dummy RegressionModel type
struct DummyMod <: RegressionModel
    beta::Vector{Float64}
    x::Matrix
    y::Vector
end

## dumb fit method: just copy the x and y input over
StatsBase.fit(::Type{DummyMod}, x::Matrix, y::Vector) =
    DummyMod(collect(1:size(x, 2)), x, y)
StatsBase.response(mod::DummyMod) = mod.y
## dumb coeftable: just prints the "beta" values
StatsBase.coeftable(mod::DummyMod) =
    CoefTable(reshape(mod.beta, (size(mod.beta,1), 1)),
              ["'beta' value"],
              ["" for n in 1:size(mod.x,2)],
              0)
# dumb predict: return values predicted by "beta" and dummy confidence bounds
function StatsBase.predict(mod::DummyMod;
                           interval::Union{Nothing,Symbol}=nothing)
    pred = mod.x * mod.beta
    if interval === nothing
        return pred
    elseif interval === :prediction
        return (prediction=pred, lower=pred .- 1, upper=pred .+ 1)
    else
        throw(ArgumentError("value not allowed for interval"))
    end
end
function StatsBase.predict(mod::DummyMod, newX::Matrix;
                           interval::Union{Nothing,Symbol}=nothing)
    pred = newX * mod.beta
    if interval === nothing
        return pred
    elseif interval === :prediction
        return (prediction=pred, lower=pred .- 1, upper=pred .+ 1)
    else
        throw(ArgumentError("value not allowed for interval"))
    end
end
StatsBase.dof(mod::DummyMod) = length(mod.beta)
StatsBase.dof_residual(mod::DummyMod) = length(mod.y) - length(mod.beta)
StatsBase.nobs(mod::DummyMod) = length(mod.y)
StatsBase.deviance(mod::DummyMod) = sum((response(mod) .- predict(mod)).^2)
# Incorrect but simple definition
StatsModels.isnested(mod1::DummyMod, mod2::DummyMod; atol::Real=0.0) =
    dof(mod1) <= dof(mod2)

# A dummy RegressionModel type that does not support intercept
struct DummyModNoIntercept <: RegressionModel
    beta::Vector{Float64}
    x::Matrix
    y::Vector
end

StatsModels.drop_intercept(::Type{DummyModNoIntercept}) = true

## dumb fit method: just copy the x and y input over
StatsBase.fit(::Type{DummyModNoIntercept}, x::Matrix, y::Vector) =
    DummyModNoIntercept(collect(1:size(x, 2)), x, y)
StatsBase.response(mod::DummyModNoIntercept) = mod.y
## dumb coeftable: just prints the "beta" values
StatsBase.coeftable(mod::DummyModNoIntercept) =
    CoefTable(reshape(mod.beta, (size(mod.beta,1), 1)),
              ["'beta' value"],
              ["" for n in 1:size(mod.x,2)],
              0)
# dumb predict: return values predicted by "beta" and dummy confidence bounds
function StatsBase.predict(mod::DummyModNoIntercept;
                           interval::Union{Nothing,Symbol}=nothing)
    pred = mod.x * mod.beta
    if interval === nothing
        return pred
    elseif interval === :prediction
        return (prediction=pred, lower=pred .- 1, upper=pred .+ 1)
    else
        throw(ArgumentError("value not allowed for interval"))
    end
end
function StatsBase.predict(mod::DummyModNoIntercept, newX::Matrix;
                           interval::Union{Nothing,Symbol}=nothing)
    pred = newX * mod.beta
    if interval === nothing
        return pred
    elseif interval === :prediction
        return (prediction=pred, lower=pred .- 1, upper=pred .+ 1)
    else
        throw(ArgumentError("value not allowed for interval"))
    end
end
StatsBase.dof(mod::DummyModNoIntercept) = length(mod.beta)
StatsBase.dof_residual(mod::DummyModNoIntercept) = length(mod.y) - length(mod.beta)
StatsBase.nobs(mod::DummyModNoIntercept) = length(mod.y)
StatsBase.deviance(mod::DummyModNoIntercept) = sum((response(mod) .- predict(mod)).^2)
# isnested not implemented to test fallback

## Another dummy model type to test fall-through show method
struct DummyModTwo <: RegressionModel
    msg::String
end

StatsBase.fit(::Type{DummyModTwo}, ::Matrix, ::Vector) = DummyModTwo("hello!")
Base.show(io::IO, m::DummyModTwo) = println(io, m.msg)

@testset "stat model types" begin

    ## Test fitting
    d = DataFrame(y = 1:4,
                  x1 = allowmissing(5:8),
                  x2 = 9:12,
                  x3 = 13:16,
                  x4 = 17:20)

    d.x1p = categorical(d.x1)

    f = @formula(y ~ x1 * x2)
    m = fit(DummyMod, f, d)
    @test response(m) == Array(d.y)

    ## coefnames delegated to model frame by default
    @test coefnames(m) == coefnames(ModelFrame(f, d)) == ["(Intercept)", "x1", "x2", "x1 & x2"]

    ## test prediction method
    ## vanilla
    @test predict(m) == [ ones(size(d,1)) Array(d.x1) Array(d.x2) Array(d.x1).*Array(d.x2) ] * collect(1:4)

    ## new data from matrix
    mm = ModelMatrix(ModelFrame(f, d))
    p = predict(m, mm.m)
    @test p == mm.m * collect(1:4)
    p2 = predict(m, mm.m, interval=:prediction)
    @test p2 isa NamedTuple
    @test p2.prediction == p
    @test p2.lower == p .- 1
    @test p2.upper == p .+ 1

    ## new data from DataFrame (via ModelMatrix)
    @test predict(m, d) == p
    p3 = predict(m, d, interval=:prediction)
    @test p3 isa DataFrame
    @test p3.prediction == p
    @test p3.lower == p .- 1
    @test p3.upper == p .+ 1

    d2 = deepcopy(d)
    d2[3, :x1] = missing
    @test length(predict(m, d2)) == 4

    ## test copying of names from Terms to CoefTable
    ct = coeftable(m)
    @test ct.rownms == ["(Intercept)", "x1", "x2", "x1 & x2"]

    ## show with coeftable defined
    io = IOBuffer()
    show(io, m)

    ## with categorical variables
    f2 = @formula(y ~ x1p)
    m2 = fit(DummyMod, f2, d)

    @test coeftable(m2).rownms == ["(Intercept)", "x1p: 6", "x1p: 7", "x1p: 8"]

    ## predict w/ new data missing levels
    @test predict(m2, d[2:4, :]) == predict(m2)[2:4]

    ## predict w/ new data with _extra_ levels (throws an error)
    d3 = deepcopy(d)
    d3[1, :x1] = 0
    d3.x1p = categorical(d3.x1)
    # TODO: check for level mismatch earlier...this throws a KeyError when it
    # goes to do the lookup in the contrasts matrix from the previously
    # generated categorical term.
    @test_throws KeyError predict(m2, d3)
    # @test_throws ArgumentError predict(m2, d3)

    ## predict with dataframe that doesn't have the dependent variable
    d4 = deepcopy(d)
    select!(d4, Not(:y))
    @test predict(m, d4) == predict(m, d)

    ## attempting to fit with d4 should fail since it doesn't have :y
    @test_throws ErrorException fit(DummyMod, f, d4)

    ## fit with contrasts specified
    d.x2p = categorical(d.x2)
    f3 = @formula(y ~ x1p + x2p)
    m3 = fit(DummyMod, f3, d)
    fit(DummyMod, f3, d, contrasts = Dict(:x1p => EffectsCoding()))
    fit(DummyMod, f3, d, contrasts = Dict(:x1p => EffectsCoding(),
                                          :x2p => DummyCoding()))
    @test_throws Exception fit(DummyMod, f3, d, contrasts = Dict(:x1p => EffectsCoding(),
                                                                 :x2p => 1))


    f1 = @formula(y ~ 1 + x1 * x2)
    f2 = @formula(y ~ 0 + x1 * x2)
    f3 = @formula(y ~ x1 * x2)
    @test_throws ArgumentError m1 = fit(DummyModNoIntercept, f1, d)
    m2 = fit(DummyModNoIntercept, f2, d)
    m3 = fit(DummyModNoIntercept, f3, d)
    ct2 = coeftable(m2)
    ct3 = coeftable(m3)
    @test ct3.rownms == ct2.rownms == ["x1", "x2", "x1 & x2"]
    @test predict(m2, d[2:4, :]) == predict(m2)[2:4]
    @test predict(m3, d[2:4, :]) == predict(m3)[2:4]

    f1 = @formula(y ~ 1 + x1p)
    f2 = @formula(y ~ 0 + x1p)
    f3 = @formula(y ~ x1p)
    @test_throws ArgumentError m1 = fit(DummyModNoIntercept, f1, d)
    m2 = fit(DummyModNoIntercept, f2, d)
    m3 = fit(DummyModNoIntercept, f3, d)
    ct2 = coeftable(m2)
    ct3 = coeftable(m3)
    @test ct2.rownms == ct3.rownms == ["x1p: 6", "x1p: 7", "x1p: 8"]
    m4 = fit(DummyModNoIntercept, f3, d, contrasts = Dict(:x1p => EffectsCoding()))
    @test predict(m2, d[2:4, :]) == predict(m2)[2:4]
    @test predict(m3, d[2:4, :]) == predict(m3)[2:4]
    @test predict(m4, d[2:4, :]) == predict(m4)[2:4]

    m2 = fit(DummyModTwo, f, d)
    show(io, m2)

end

@testset "lrtest" begin

    y = collect(1:4)
    x1 = 2:5
    x2 = [1, 5, 3, 1]

    m0 = DummyMod([1], ones(4, 1), y)
    m1 = DummyMod([1, 0.3], [ones(4, 1) x1], y)
    m2 = DummyMod([1, 0.25, 0.05, 0.04], [ones(4, 1) x1 x2 x1.*x2], y)

    @test_throws ArgumentError lrtest(m0)
    @test_throws ArgumentError lrtest(m0, m0)
    @test_throws ArgumentError lrtest(m0, m2, m1)
    @test_throws ArgumentError lrtest(m1, m0, m2)
    @test_throws ArgumentError lrtest(m2, m0, m1)

    m1b = DummyMod([1, 0.3], [ones(3, 1) x1[2:end]], y[2:end])
    @test_throws ArgumentError lrtest(m0, m1b)

    lr1 = lrtest(m0, m1)
    @test isnan(lr1.pval[1])
    @test lr1.pval[2] ≈ 0.0010484433450981662
    @test sprint(show, lr1) == """
        Likelihood-ratio test: 2 models fitted on 4 observations
        ──────────────────────────────────────────────
             DOF  ΔDOF  Deviance  ΔDeviance  p(>Chisq)
        ──────────────────────────────────────────────
        [1]    1         14.0000                      
        [2]    2     1    3.2600   -10.7400     0.0010
        ──────────────────────────────────────────────"""

    m0 = DummyModNoIntercept(Float64[], ones(4, 0), y)
    m1 = DummyModNoIntercept([0.3], reshape(x1, :, 1), y)
    m2 = DummyModNoIntercept([0.25, 0.05, 0.04], [x1 x2 x1.*x2], y)

    @test_throws ArgumentError lrtest(m0)
    @test_throws ArgumentError lrtest(m0, m0)
    @test_throws ArgumentError lrtest(m0, m2, m1)
    @test_throws ArgumentError lrtest(m1, m0, m2)
    @test_throws ArgumentError lrtest(m2, m0, m1)

    m1b = DummyModNoIntercept([0.3], reshape(x1[2:end], :, 1), y[2:end])
    @test_throws ArgumentError lrtest(m0, m1b)

    # Incorrect, but check that it doesn't throw an error
    lr2 = @test_logs((:warn, "Could not check whether models are nested " *
                     "as model type DummyModNoIntercept does not implement isnested: " *
                     "results may not be meaningful"),
                     lrtest(m0, m1))
    @test isnan(lr2.pval[1])
    @test lr2.pval[2] ≈ 1.2147224767092312e-5

    # in 1.6, p value printing has changed (JuliaStats/StatsBase.jl#606)
    if VERSION > v"1.6.0-DEV"
        @test sprint(show, lr2) == """
            Likelihood-ratio test: 2 models fitted on 4 observations
            ──────────────────────────────────────────────
                 DOF  ΔDOF  Deviance  ΔDeviance  p(>Chisq)
            ──────────────────────────────────────────────
            [1]    0         30.0000                      
            [2]    1     1   10.8600   -19.1400     <1e-04
            ──────────────────────────────────────────────"""
    else
        @test sprint(show, lr2) == """
            Likelihood-ratio test: 2 models fitted on 4 observations
            ──────────────────────────────────────────────
                 DOF  ΔDOF  Deviance  ΔDeviance  p(>Chisq)
            ──────────────────────────────────────────────
            [1]    0         30.0000                      
            [2]    1     1   10.8600   -19.1400      <1e-4
            ──────────────────────────────────────────────"""
    end
end
